/*
 * KMap.cpp
 *
 *  Created on: Sep 22, 2008
 *      Author: Santi Ontan
 */

#ifdef KITSCHY_DEBUG_MEMORY
#include "debug_memorymanager.h"
#endif

#ifdef _WIN32
#include "windows.h"
#endif

#include "math.h"
#include "stdio.h"

#include "SDL.h"
#include "SDL_image.h"


#include "GLTile.h"
#include "debug.h"
#include "XMLparser.h"

#include "Knightmare.h"
#include "KMap.h"

#include "KObject.h"
#include "KTile.h"
#include "Column.h"
#include "ColumnComplete.h"
#include "Blob.h"
#include "Bat.h"
#include "Satyr.h"

#include "KObjectCreator.h"

#include "KnightmareCtnt.h"

KMap::KMap(char *mapName, char *mapWavesName)
{
	SDL_Surface *tileset_sfc = 0;
	FILE *fp = fopen(mapName,"r+");
	int ntx = 0,nty = 0;
	int i;

	m_tileset_n_tiles = 0;
	m_tileset_tiles = 0;

	m_cycle = 0;

	output_debug_message("Loading map: '%s'\n",mapName);

	if (fp!=0) {
		XMLNode *map = XMLNode::from_file(fp);

		if (map!=0) {
			// Create the empty map:
			m_tile_width = atoi(map->get_attribute("tilewidth")->get());
			m_tile_height = atoi(map->get_attribute("tileheight")->get());
			m_width = atoi(map->get_attribute("width")->get())*m_tile_width;
			m_height = atoi(map->get_attribute("height")->get())*m_tile_height;

			m_grid_width = m_width/4;
			m_grid_height = m_height/32;

			m_grid_dimensions_x = 4;
			m_grid_dimensions_y = 40;

			m_CG = new GridCell *[m_grid_dimensions_y*m_grid_dimensions_x];
			for(i=0;i<m_grid_dimensions_x*m_grid_dimensions_y;i++) {
				m_CG[i] = new GridCell();
			}

			// Load the tile set:
			{
				XMLNode *tileset = map->get_children("tileset");
				if (tileset!=0) {
					char tileset_image[256];
					output_debug_message("The tiles are of size %ix%i\n",m_tile_width,m_tile_height);

					sprintf(tileset_image,"graphics/%s",tileset->get_children("image")->get_attribute("source")->get());
					tileset_sfc = IMG_Load(tileset_image);

					if (tileset_sfc!=0) {
						output_debug_message("Tile set image load properly. '%s'\n",tileset_image);

						ntx = tileset_sfc->w/m_tile_width;
						nty = tileset_sfc->h/m_tile_height;

						int i,j;

						m_tileset_tiles = new GLTile *[nty*ntx];
						m_tileset_n_tiles = nty*ntx;

						for(i = 0;i<nty;i++) {
							for(j = 0;j<ntx;j++) {
								m_tileset_tiles[j+i*ntx] = 0;
							}
						}
					} else {
						output_debug_message("Cannot load the tile set image! '%s'\n",tileset_image);
					}
				} else {
					output_debug_message("No tile set found in the map!\n");
				}
			}

			// Load the tile layers:
			{
				int layer_count = 0;
				int number_load = 0;
				List<XMLNode> *l;
				XMLNode *layer;

				l = map->get_children();
				l->Rewind();
				while(l->Iterate(layer)) {
					if (layer->get_type()->cmp("layer")) {
						int width,height;
						int count = 0,tile_id;

						output_debug_message("Loading layer %i.\n",layer_count);

						width = atoi(layer->get_attribute("width")->get());
						height = atoi(layer->get_attribute("height")->get());

						output_debug_message("size of layer is %ix%i.\n",width,height);

						List<XMLNode> *l2;
						XMLNode *tile;

						l2 = layer->get_children("data")->get_children();
						l2->Rewind();
						while(l2->Iterate(tile)) {
							tile_id = atoi(tile->get_attribute("gid")->get());

							if (tile_id>0) {
								tile_id--;

								if (tile_id>=ntx*nty) {
									output_debug_message("Tile out of range! %i.\n",tile_id);
									tile_id = 0;
								}

								if (m_tileset_tiles[tile_id]==0) {
									int i,j;
									j = tile_id%ntx;
									i = tile_id/ntx;
									m_tileset_tiles[tile_id] = new GLTile(tileset_sfc,j*m_tile_width,i*m_tile_height,m_tile_width,m_tile_height);
									output_debug_message("Loading tile %i (%i,%i) - (%i,%i) from image %p, of %i,%i.\n",tile_id,
														 j*m_tile_width,i*m_tile_height,m_tile_width,m_tile_height,
														 tileset_sfc,tileset_sfc->w,tileset_sfc->h);
									number_load++;
								}

								addPassiveObject(layer_count,new KTile((count%width)*m_tile_width,(count/width)*m_tile_height,m_tileset_tiles[tile_id]));
							}

							count++;
						}
					}
					layer_count++;
				}

				output_debug_message("%i tiles used and load.\n",number_load);

			}

			// Load the tile layers:
			{
				List<XMLNode> *l;
				XMLNode *layer;

				l = map->get_children();
				l->Rewind();
				while(l->Iterate(layer)) {
					if (layer->get_type()->cmp("objectgroup")) {

						List<XMLNode> *l2;
						XMLNode *object;

						l2 = layer->get_children();
						l2->Rewind();
						while(l2->Iterate(object)) {
							if (object->get_attribute("name")->cmp("pillar")) {
								addObject(4,new Column(atoi(object->get_attribute("x")->get())+40,atoi(object->get_attribute("y")->get())+40,-40));
							}
							if (object->get_attribute("name")->cmp("pillar-complete")) {
								addObject(4,new ColumnComplete(atoi(object->get_attribute("x")->get())+40,atoi(object->get_attribute("y")->get())+40,-60));
							}
						}
					}
				}
			}

			delete map;

		} else {
			output_debug_message("Error loading map!\n");
		}

		if (tileset_sfc!=0) SDL_FreeSurface(tileset_sfc);

		fclose(fp);
	}

	output_debug_message("Loading waves: '%s'\n",mapWavesName);
	fp = fopen(mapWavesName,"r+");

	if (fp!=0) {
		XMLNode *waves = XMLNode::from_file(fp);

		List<XMLNode> *l;
		XMLNode *wave;

		l = waves->get_children();
		l->Rewind();
		while(l->Iterate(wave)) {
			// Process each of the waves:
			Symbol *type = 0;
			int n = -1;

			if (wave->get_attribute("entity")!=0) type = new Symbol(wave->get_attribute("entity"));
			if (wave->get_attribute("number")!=0) n = atoi(wave->get_attribute("number")->get());
			
			m_waves.Add(new EnemyWave(type,n));
		}


		fclose(fp);
	}

    m_current_y = (float)m_height-SCREEN_Y;

	m_plane = new K3DObject(m_width/2,m_height/2,0,0);
	m_plane->createPlane(m_width/2,m_height/2,1,1,1,0);

}


KMap::~KMap()
{
	int i;

	delete m_plane;
	m_plane = 0;

	for(i=0;i<m_tileset_n_tiles;i++) {
		if (m_tileset_tiles[i]!=0) {
			delete m_tileset_tiles[i];
			m_tileset_tiles[i]=0;
		}
	}
	delete []m_tileset_tiles;



	for(i=0;i<m_grid_dimensions_x*m_grid_dimensions_y;i++) {
		delete m_CG[i];
		m_CG[i]=0;
	}

	for(i=0;i<NLAYERS;i++) m_active_objects[i].ExtractAll();

	m_characters.ExtractAll();

}


void KMap::addObject(int layer,KObject *o) {
	GridCell *cell = m_CG[toGrid(int(o->get_x()),int(o->get_y()))];
	cell->m_objects[layer].Add(o);
	m_objects[layer].Add(o);

	if (o->is_a(kplayer_symbol)) {
		m_characters.Add(o);
	}
	
	o->set_layer(layer);
}


void KMap::addActiveObject(int layer,KObject *o) {
	GridCell *cell = m_CG[toGrid(int(o->get_x()),int(o->get_y()))];
	cell->m_objects[layer].Add(o);
	m_objects[layer].Add(o);
	m_active_objects[layer].Add(o);
	
	if (o->is_a(kplayer_symbol)) {
		m_characters.Add(o);
	}
	
	o->set_layer(layer);
}


void KMap::addPassiveObject(int layer,KObject *o) {
	GridCell *cell = m_CG[toGrid(int(o->get_x()),int(o->get_y()))];
	cell->m_passive_objects[layer].Add(o);
	m_passive_objects[layer].Add(o);

	o->set_layer(layer);
}

void KMap::deleteObject(KObject *o) {
	int l = o->get_layer();
	m_objects[l].DeleteElement(o);
	m_active_objects[l].DeleteElement(o);
	GridCell *cell = m_CG[toGrid(int(o->get_x()),int(o->get_y()))];
	cell->m_objects[l].DeleteElement(o);

	if (o->is_a(kplayer_symbol)) {
		m_characters.DeleteElement(o);
	}
}

void KMap::deletePassiveObject(KObject *o) {
	int l = o->get_layer();
	m_passive_objects[l].DeleteElement(o);
	GridCell *cell = m_CG[toGrid(int(o->get_x()),int(o->get_y()))];
	cell->m_passive_objects[l].DeleteElement (o);
}


void KMap::toGrid(int x,int y,int *gx,int *gy) {
	*gx = x/m_grid_width;
	*gy = y/m_grid_height;

}

int KMap::toGrid(int x,int y) {
	return (x/m_grid_width)+ (y/m_grid_height)*m_grid_dimensions_x;
}

void KMap::draw(int x,int y,float zoom,float angle,GLTManager *GLTM,bool preserve_angle)
{
	int i,j,layer;
	List<KObject> l;
	KObject *o;
	List<K3DObject> l2;	// Here we will add all the 3D objects that require shadows

	y+=m_current_y;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glTranslatef(SCREEN_X/2,SCREEN_Y/2,0);

	int firsty=0,lasty=m_grid_dimensions_y;

	lasty = ((y + SCREEN_Y)/m_grid_height)+1;
	if (angle!=0) lasty++;
	if (lasty>m_grid_dimensions_y) lasty = m_grid_dimensions_y;
	firsty = lasty-((SCREEN_Y/m_grid_height)+4);
	if (angle!=0) firsty-=4;
	if (firsty<0) firsty = 0;

	glScalef(zoom,zoom,zoom);
	glRotatef(-angle,1,0,0);

	for(layer=0;layer<NLAYERS;layer++) {
		for(i=firsty;i<lasty;i++) {
			for(j=0;j<m_grid_dimensions_x;j++) {

				l.Instance(m_CG[i*m_grid_dimensions_x + j]->m_passive_objects[layer]);
				l.Rewind();
				while(l.Iterate(o)) {
					o->draw((preserve_angle ? angle:0),-(x+SCREEN_X/2),-(y+SCREEN_Y/2),GLTM);
					if (o->is_a(k3dobject_symbol)) l2.Add((K3DObject *)o);
				}

				l.Instance(m_CG[i*m_grid_dimensions_x + j]->m_objects[layer]);
				l.Rewind();
				while(l.Iterate(o)) {
					o->draw((preserve_angle ? angle:0),-(x+SCREEN_X/2),-(y+SCREEN_Y/2),GLTM);
					if (o->is_a(k3dobject_symbol)) l2.Add((K3DObject *)o);
				}
			}
		}
	}

	m_plane->draw((preserve_angle ? angle:0),-(x+SCREEN_X/2),-(y+SCREEN_Y/2),GLTM);

	l2.Add(m_plane);
	{
		Vector light_position;
		light_position.x=-20000;
		light_position.y=-15000;
		light_position.z=-55000;
		drawShadows(&l2,-(x+SCREEN_X/2),-(y+SCREEN_Y/2),light_position.x,light_position.y,light_position.z,100000);
	}
	l2.ExtractAll();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

}


void KMap::drawShadows(List<K3DObject> *objects,int x,int y,float vlx,float vly,float vlz,float depth)
{
	K3DObject *o;

//	output_debug_message("drawing shadows with %i objects\n",objects->Length());

	glPushAttrib( GL_ALL_ATTRIB_BITS );

	glEnable(GL_DEPTH_TEST);
	glDisable( GL_LIGHTING );
	glDepthMask( GL_FALSE );
	glDepthFunc( GL_LEQUAL );
	glEnable( GL_STENCIL_TEST );
	glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
	glStencilFunc( GL_ALWAYS, 1, 0xFFFFFFFFL );

	glMatrixMode(GL_MODELVIEW);

	glEnable(GL_CULL_FACE);
	glFrontFace( GL_CCW );
	glStencilOp( GL_KEEP, GL_KEEP, GL_INCR );
	objects->Rewind();
	while(objects->Iterate(o)) {
		o->drawShadows(x,y,vlx,vly,vlz,depth);
	}
	glFrontFace( GL_CW );
	glStencilOp( GL_KEEP, GL_KEEP, GL_DECR );
	objects->Rewind();
	while(objects->Iterate(o)) {
		o->drawShadows(x,y,vlx,vly,vlz,depth);
	}
	glFrontFace( GL_CCW );
	glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );

	glDepthFunc(GL_ALWAYS);
	glDisable(GL_CULL_FACE);
	glColor4f(0.0,0.0,0.0,0.5f);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glStencilFunc(GL_NOTEQUAL,0,0xFFFFFFFFL);
	glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glBegin(GL_TRIANGLE_STRIP);
	glVertex3f(-1000.0, 1000.0,0.0);
	glVertex3f(-1000.0,-1000.0,0.0);
	glVertex3f( 1000.0, 1000.0,0.0);
	glVertex3f( 1000.0,-1000.0,0.0);
	glEnd();
	glStencilFunc(GL_ALWAYS,0,0);
	glPopMatrix();

	glPopAttrib();
}


KObject *KMap::collisionWithType(KObject *o, float offs_x, float offs_y, Symbol *objectType)
{
	List<Symbol> l;
	l.Add(objectType);
	KObject *result = collisionWithTypes(o,offs_x,offs_y,&l);
	l.ExtractAll();
	return result;
}


KObject *KMap::collisionWithTypes(KObject *o, float offs_x, float offs_y, List<Symbol> *objectTypes)
{
	int grid_tolerance = 2;
	int gx = 0, gy = 0;
	int x,y,l;
	
	toGrid(o->get_x()+offs_x,o->get_y()+offs_y,&gx,&gy);
	
	for(y = gy-grid_tolerance; y<gy+grid_tolerance ; y++) {
		if (y>=0 && y<m_grid_dimensions_y) {
			for(x = gx-grid_tolerance; x<gx+grid_tolerance ; x++) {
				if (x>=0 && x<m_grid_dimensions_x) {
					GridCell *gc = m_CG[x+y*m_grid_dimensions_x];
					for(l = 0;l<NLAYERS;l++) {
						List<KObject> list;
						List<Symbol> list2;
						KObject *o2;
						Symbol *s;
						bool found;
						
						list2.Instance(*objectTypes);
						list.Instance(gc->m_objects[l]);
						list.Rewind();
						while(list.Iterate(o2)) {
							found = false;
							list2.Rewind();
							while(!found && list2.Iterate(s)) {
								if (o2->is_a(s)) found = true;
							}
//							if (found) output_debug_message("AO: going to test collision between %s and %s\n",o->get_class()->get(),o2->get_class()->get());
							if (found && o!=o2 && o->collision(offs_x, offs_y, o2)) {
//								output_debug_message("Collision between %s and %s",o->get_class()->get(),o2->get_class()->get());
								return o2;
							}
						}
						/*
						list.Instance(gc->m_passive_objects[l]);
						list.Rewind();
						while(list.Iterate(o2)) {
							found = false;
							list2.Rewind();
							while(!found && list2.Iterate(s)) {
								if (o2->is_a(s)) found = true;
							}
							if (found) output_debug_message("PO: going to test collision between %s and %s\n",o->get_class()->get(),o2->get_class()->get());
							if (found && o!=o2 && o->collision(offs_x, offs_y, o2)) {
//								output_debug_message("Collision between %s and %s",o->get_class()->get(),o2->get_class()->get());
								return true;
							}
						}
						*/
					}
				}			
			}
		}
	}
								
	return 0;
}


KObject *KMap::collision(KObject *o, float offs_x, float offs_y) 
{
	int gx = 0, gy = 0;
	int x,y,l;
	
	toGrid(o->get_x()+offs_x,o->get_y()+offs_y,&gx,&gy);
	
	for(y = gy-1; y<gy+1 ; y++) {
		if (y>=0 && y<m_grid_dimensions_y) {
			for(x = gx-1; x<gx+1 ; x++) {
				if (x>=0 && x<m_grid_dimensions_x) {
					GridCell *gc = m_CG[x+y*m_grid_dimensions_x];
					for(l = 0;l<NLAYERS;l++) {
						List<KObject> list;
						KObject *o2;
						list.Instance(gc->m_objects[l]);
						while(list.Iterate(o2)) {
							if (o->collision(offs_x, offs_y, o2)) return o2;
						}
						/*
						list.Instance(gc->m_passive_objects[l]);
						while(list.Iterate(o2)) {
							if (o->collision(offs_x, offs_y, o2)) return o2;
						}
						*/
					}
				}			
			}
		}
	}
	
	return 0;
}


float KMap::get_current_y() {
	return m_current_y;
}


int KMap::get_width() {
	return m_width;
}


List<KObject> *KMap::getObjectsByType(Symbol *type) {
	List<KObject> *res = new List<KObject>();
	List<KObject> l;
	KObject *o;
	
	for(int layer=0;layer<NLAYERS;layer++) {
		l.Instance(m_active_objects[layer]);
		l.Rewind();
		while(l.Iterate(o)) {
			if (o->is_a(type)) res->Add(o);
		}
	}
	
	return res;
}



bool KMap::cycle(List<VirtualController> *vcl, Knightmare *game, GLTManager *GLTM, SFXManager *SFXM)
{
	int layer;
	List<KObject> l;
	KObject *o;
	List<KObject> toDelete;

	for(layer=0;layer<NLAYERS;layer++) {
		l.Instance(m_objects[layer]);
		l.Rewind();
		while(l.Iterate(o)) {
			float old_x = o->get_x();
			float old_y = o->get_y();

			if (!o->cycle(vcl,game, GLTM, SFXM)) {
				toDelete.Add(o);
			}

			if (m_current_y>0 && (m_cycle%4)==0) {
				if (m_characters.MemberRefP(o)) {
					o->set_y(o->get_y()-1.0f);
				}
			}

			// If the object moves, reposition them in the right cells:
			if (o->get_x()!=old_x || o->get_y()!=old_y) {
				int cell=toGrid(old_x,old_y);
				m_CG[cell]->m_objects[layer].DeleteElement(o);
				cell=toGrid(o->get_x(),o->get_y());
				m_CG[cell]->m_objects[layer].Add(o);
			}
		}
	}
	
	while(!toDelete.EmptyP()) {
		deleteObject(toDelete.ExtractIni());
	}

	// Check if new enemies have to be spawned:
	{
		EnemyWave *ew;

		// look for enemies in the map:
		List<KObject> *l = getObjectsByType(kenemy_symbol);
		bool enemies_dead = false;
		if (l==0 || l->Length()==0) enemies_dead = true;

		// next wave:
		ew = m_waves[0];
		if (ew!=0) {
			if (ew->m_type->cmp("long-wait")) {
				if (enemies_dead) {
					ew->m_number++;
					if (ew->m_number>200) {
#ifdef __DEBUG_MESSAGES
						output_debug_message("ENEMYWAVE: finishing long pause %i with currrent_y %g (cycle %i)\n",ew->m_number,get_current_y(),m_cycle);
#endif
						m_waves.ExtractIni();
						delete ew;
					}
				}
			} else if (ew->m_type->cmp("wait")) {
				if (enemies_dead) {
					ew->m_number++;
					if (ew->m_number>100) {
#ifdef __DEBUG_MESSAGES
						output_debug_message("ENEMYWAVE: finishing pause %i with currrent_y %g (cycle %i)\n",ew->m_number,get_current_y(),m_cycle);
#endif
						m_waves.ExtractIni();
						delete ew;
					}
				}
			} else if (ew->m_type->cmp("blob")) {
				// spawn some blobs:
				if (enemies_dead) {
#ifdef __DEBUG_MESSAGES
					output_debug_message("ENEMYWAVE: Spawning %i blobs with currrent_y %g (cycle %i)\n",ew->m_number,get_current_y(),m_cycle);
#endif
					for(int i = 0;i<ew->m_number;i++) {
						addActiveObject(2,new Blob(m_tile_width*5 + (rand()%22)*m_tile_width,get_current_y()-(i+1)*m_tile_height*3,game->get_sfx_volume(),GLTM));
					}
					m_waves.ExtractIni();
					delete ew;
				}
			} else if (ew->m_type->cmp("bats-side-to-side")) {
				// spawn some bats:
				if (ew->m_number!=5 || enemies_dead) {
#ifdef __DEBUG_MESSAGES
					output_debug_message("ENEMYWAVE: Spawning one bat-side-to-side with currrent_y %g (cycle %i)\n",get_current_y(),m_cycle);
#endif
					if ((m_cycle%16)==0) {
						addActiveObject(2,new Bat(m_tile_width*16,get_current_y()-3*m_tile_height,game->get_sfx_volume(),GLTM));
						ew->m_number--;
						if (ew->m_number<=0) {
							m_waves.ExtractIni();
							delete ew;
						}
					}
				}
			}
		}
	}

	if (m_current_y>0 && (m_cycle%4)==0) {
		m_current_y-=1.0f;
	}

	m_cycle++;

	return true;
}

